import '../../stencil.core';
import { EventEmitter } from '../../stencil.core';
import { EditController } from './editController';
import { IRegion } from './interfaces';
export declare class HexEditor {
    editController: EditController;
    regionScaleWidth: number;
    regionScaleHeight: number;
    /**
     * contains metadata of the given file
     * @type {File}
     * @memberof HexEditor
     */
    fileMetadata: File;
    /**
     * the loaded file
     *
     * @type {Uint8Array}
     * @memberof HexEditor
     */
    file: Uint8Array;
    lineNumber: number;
    tempSelection: number;
    selection: {
        start: number;
        end: number;
    };
    cursor: number;
    /**
     * weather or not to display ASCII on the side
     *
     * @type {boolean}
     * @memberof HexEditor
     */
    displayAscii: boolean;
    /**
     * the number of lines to display at once
     *
     * @type {number}
     * @memberof HexEditor
     */
    maxLines: number;
    /**
     * the number of bytes to display per line
     *
     * @type {number}
     * @memberof HexEditor
     */
    bytesPerLine: number;
    /**
     * currently does nothing
     * it WOULD force a line break every X bytes
     * @type {number}
     * @memberof HexEditor
     * @deprecated
     */
    bytesUntilForcedLine: number;
    /**
     * weather or not to replace typical ASCII values
     * with their ASCII value representation
     * ( ex: 0x61 ==> ".a" )
     *
     * @type {boolean}
     * @memberof HexEditor
     */
    asciiInline: boolean;
    /**
     * the number of bytes between separators
     *
     * @type {number}
     * @memberof HexEditor
     */
    bytesPerGroup: number;
    /**
     * the mode of operation:
     * region:
     *    used to highlight different regions. Hovering over
     *    a region displays a tooltip
     * edit:
     *    regions are displayed in the background, allowing
     *    the user to edit directly
     * noregion:
     *    regions are not displayed at all
     *
     * @type {("region" | "edit" | "noregion")}
     * @memberof HexEditor
     */
    mode: "region" | "edit" | "noregion";
    /**
     * the mode of data entry:
     * insert:
     *    inserts data between bytes
     * overwrite:
     *    overwrites the currently selected byte
     * readonly:
     *    no edits are possible
     *
     * @type {("insert" | "overwrite" | "readonly")}
     * @memberof HexEditor
     */
    editType: "insert" | "overwrite" | "readonly";
    /**
     * the number of regions to traverse
     *
     * @type {number}
     * @memberof HexEditor
     */
    regionDepth: number;
    /**
     * the region data. Data will be displayed in the tooltip
     * if mode is set to "region"
     *
     * @type {IRegion[]}
     * @memberof HexEditor
     */
    regions: IRegion[];
    /**
     * Emitted when the lineNumber changes
     *
     * @type {EventEmitter}
     * @memberof HexEditor
     */
    hexLineChanged: EventEmitter;
    /**
     * Emitted on the change of the cursor's position
     *
     * @type {EventEmitter}
     * @memberof HexEditor
     */
    hexCursorChanged: EventEmitter;
    /**
     * Emitted when the selection changes
     *
     * @type {EventEmitter}
     * @memberof HexEditor
     */
    hexSelectionChanged: EventEmitter;
    /**
     * fired when the file's data changes
     *
     * @type {EventEmitter}
     * @memberof HexEditor
     */
    hexDataChanged: EventEmitter;
    /**
     * fired when the component loads
     */
    hexLoaded: EventEmitter;
    componentWillLoad(): void;
    componentDidLoad(): void;
    /**
    * accepts and reads the given file, storing the result in
    * the file variable
    * @param file
    */
    acceptFile(file: File): Promise<void>;
    /**
     * returns the edited file
     *
     * @returns {(Promise<Uint8Array | void>)}
     * @memberof HexEditor
     */
    saveFile(): Promise<Uint8Array | void>;
    /**
     * sets the line number
     *
     * @param {number} newLineNumber
     * @memberof HexEditor
     */
    setLineNumber(newLineNumber: number): Promise<void>;
    /**
     * sets the new cursor position
     *
     * @param {number} newCursorPosition
     * @memberof HexEditor
     */
    setCursorPosition(newCursorPosition: number): Promise<void>;
    /**
     * sets the new selection bounds.
     * @param {{start?: number, end?: number}} newSelection
     * @memberof HexEditor
     */
    setSelection(newSelection: {
        start?: number;
        end?: number;
    }): Promise<void>;
    /**
     * builds the elements responsible for the hex view
     */
    buildHexView(): {
        lineViews: any[];
        charViews: any[];
        lineLabels: any[];
        regionMarkers: JSX.Element;
    };
    edit(evt: KeyboardEvent): void;
    /**
     * displays the full hexidecimal view
     */
    showHex(): JSX.Element;
    /**
     * gets the exact position of
     * @param evt the mousedown event
     */
    beginSelection(evt: any): void;
    endSelection(evt: any): void;
    /**
     * This must be an arrow function so it retains the reference to
     * "this" while also not being anonymous. This allows it to be
     * added as an eventlistener directly while retaining the ability
     * to remove it.
     *
     * @memberof MyComponent
     */
    scroll: (evt: WheelEvent) => void;
    render(): JSX.Element;
    _toggleScrollListener(evt: MouseEvent): void;
}
